/*
    Copyright (C) 2006 by OXIM TEAM

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/      

#include <X11/Xlib.h>
#include <X11/keysym.h>
#include "oximint.h"
#include "module.h"

typedef struct xkeymap_s {
    KeySym xkey;
    uch_t uch;
} xkeymap_t; 

static xkeymap_t fullchar[] = {		/* sorted by ascii code */
	{XK_space,		{"　"}},	/*     */
	{XK_exclam,		{"！"}},	/*  !  */
	{XK_quotedbl,		{"”"}},	/*  "  */
	{XK_numbersign,		{"＃"}},	/*  #  */
	{XK_dollar,		{"＄"}},	/*  $  */
	{XK_percent,		{"％"}},	/*  %  */
	{XK_ampersand,		{"＆"}},	/*  &  */
	{XK_apostrophe,		{"’"}},	/*  '  */
	{XK_parenleft,		{"（"}},	/*  (  */
	{XK_parenright,		{"）"}},	/*  )  */
	{XK_asterisk,		{"＊"}},	/*  *  */
	{XK_plus,		{"＋"}},	/*  +  */
	{XK_comma,		{"，"}},	/*  ,  */
	{XK_minus,		{"－"}},	/*  -  */
	{XK_period,		{"．"}},	/*  .  */
	{XK_slash,		{"╱"}},	/*  /  (undefined) */
	{XK_0,			{"０"}},	/*  0  */
	{XK_1,			{"１"}},	/*  1  */
	{XK_2,			{"２"}},	/*  2  */
	{XK_3,			{"３"}},	/*  3  */
	{XK_4,			{"４"}},	/*  4  */
	{XK_5,			{"５"}},	/*  5  */
	{XK_6,			{"６"}},	/*  6  */
	{XK_7,			{"７"}},	/*  7  */
	{XK_8,			{"８"}},	/*  8  */
	{XK_9,			{"９"}},	/*  9  */
	{XK_colon,		{"："}},	/*  :  */
	{XK_semicolon,		{"；"}},	/*  ;  */
	{XK_less,		{"＜"}},	/*  <  */
	{XK_equal,		{"＝"}},	/*  =  */
	{XK_greater,		{"＞"}},	/*  >  */
	{XK_question,		{"？"}},	/*  ?  */
	{XK_at,			{"＠"}},	/*  @  */
	{XK_A,			{"Ａ"}},	/*  A  */
	{XK_B,			{"Ｂ"}},	/*  B  */
	{XK_C,			{"Ｃ"}},	/*  C  */
	{XK_D,			{"Ｄ"}},	/*  D  */
	{XK_E,			{"Ｅ"}},	/*  E  */
	{XK_F,			{"Ｆ"}},	/*  F  */
	{XK_G,			{"Ｇ"}},	/*  G  */
	{XK_H,			{"Ｈ"}},	/*  H  */
	{XK_I,			{"Ｉ"}},	/*  I  */
	{XK_J,			{"Ｊ"}},	/*  J  */
	{XK_K,			{"Ｋ"}},	/*  K  */
	{XK_L,			{"Ｌ"}},	/*  L  */
	{XK_M,			{"Ｍ"}},	/*  M  */
	{XK_N,			{"Ｎ"}},	/*  N  */
	{XK_O,			{"Ｏ"}},	/*  O  */
	{XK_P,			{"Ｐ"}},	/*  P  */
	{XK_Q,			{"Ｑ"}},	/*  Q  */
	{XK_R,			{"Ｒ"}},	/*  R  */
	{XK_S,			{"Ｓ"}},	/*  S  */
	{XK_T,			{"Ｔ"}},	/*  T  */
	{XK_U,			{"Ｕ"}},	/*  U  */
	{XK_V,			{"Ｖ"}},	/*  V  */
	{XK_W,			{"Ｗ"}},	/*  W  */
	{XK_X,			{"Ｘ"}},	/*  X  */
	{XK_Y,			{"Ｙ"}},	/*  Y  */
	{XK_Z,			{"Ｚ"}},	/*  Z  */
	{XK_bracketleft,	{"〔"}},	/*  [  */
	{XK_backslash,		{"╲"}},	/*  \  (undefined) */
	{XK_bracketright,	{"〕"}},	/*  ]  */
	{XK_asciicircum,	{"︿"}},	/*  ^  */
	{XK_underscore,		{"ˍ"}},	/*  _  (undefined) */
	{XK_grave,		{"‵"}},	/*  `  */
	{XK_a,			{"ａ"}},	/*  a  */
	{XK_b,			{"ｂ"}},	/*  b  */
	{XK_c,			{"ｃ"}},	/*  c  */
	{XK_d,			{"ｄ"}},	/*  d  */
	{XK_e,			{"ｅ"}},	/*  e  */
	{XK_f,			{"ｆ"}},	/*  f  */
	{XK_g,			{"ｇ"}},	/*  g  */
	{XK_h,			{"ｈ"}},	/*  h  */
	{XK_i,			{"ｉ"}},	/*  i  */
	{XK_j,			{"ｊ"}},	/*  j  */
	{XK_k,			{"ｋ"}},	/*  k  */
	{XK_l,			{"ｌ"}},	/*  l  */
	{XK_m,			{"ｍ"}},	/*  m  */
	{XK_n,			{"ｎ"}},	/*  n  */
	{XK_o,			{"ｏ"}},	/*  o  */
	{XK_p,			{"ｐ"}},	/*  p  */
	{XK_q,			{"ｑ"}},	/*  q  */
	{XK_r,			{"ｒ"}},	/*  r  */
	{XK_s,			{"ｓ"}},	/*  s  */
	{XK_t,			{"ｔ"}},	/*  t  */
	{XK_u,			{"ｕ"}},	/*  u  */
	{XK_v,			{"ｖ"}},	/*  v  */
	{XK_w,			{"ｗ"}},	/*  w  */
	{XK_x,			{"ｘ"}},	/*  x  */
	{XK_y,			{"ｙ"}},	/*  y  */
	{XK_z,			{"ｚ"}},	/*  z  */
	{XK_braceleft,		{"｛"}},	/*  {  */
	{XK_bar,		{"｜"}},	/*  |  */
	{XK_braceright,		{"｝"}},	/*  }  */
	{XK_asciitilde,		{"～"}},	/*  ~  */
	{0L,			{{'\0'}}}
};

static char cch[UCH_SIZE+1];

char *
fullchar_keystroke(inpinfo_t *inpinfo, KeySym keysym)
{
    xkeymap_t *fc=fullchar;

    while (fc->xkey != 0) {
	if (keysym == fc->xkey) {
	    strncpy(cch, (char *)fc->uch.s, UCH_SIZE);
	    cch[UCH_SIZE] = '\0';
	    return cch;
	}
	fc ++;
    }
    return NULL;
}

char *
fullchar_keystring(int ch)
{
    int i;

    if ((i = ch - (int)' ') >= 95 || i < 0)
	return NULL;
    strncpy(cch, (char *)fullchar[i].uch.s, UCH_SIZE);
    cch[UCH_SIZE] = '\0';
    return cch;
}

char *
fullchar_ascii(inpinfo_t *inpinfo, int mode, keyinfo_t *keyinfo)
{
    int i;

    if (keyinfo->keystr_len != 1)
	return fullchar_keystroke(inpinfo, keyinfo->keysym);

    if ((i = (int)(keyinfo->keystr[0]-' ')) >= 95 || i < 0)
	return NULL;

    if (mode) {
	if ((keyinfo->keystate & LockMask) && (keyinfo->keystate & ShiftMask))
	    i = toupper(keyinfo->keystr[0]) - ' ';
	else
	    i = tolower(keyinfo->keystr[0]) - ' ';
    }
    strncpy(cch, (char *)fullchar[i].uch.s, UCH_SIZE);
    cch[UCH_SIZE] = '\0';
    return cch;
}

char *
halfchar_ascii(inpinfo_t *inpinfo, int mode, keyinfo_t *keyinfo)
{
    int i;

    if (keyinfo->keystr_len != 1)
	return NULL;

    if ((i = (int)(keyinfo->keystr[0]-' ')) >= 95 || i < 0)
	return NULL;

    if (mode) {
	if ((keyinfo->keystate & LockMask) && (keyinfo->keystate & ShiftMask))
	    i = toupper(keyinfo->keystr[0]);
	else
	    i = tolower(keyinfo->keystr[0]);

        cch[0] = (char)i;
        cch[1] = '\0';
        return cch;
    }
    else
	return NULL;
}

KeySym
keysym_ascii(int ch)
{
    int i;

    if ((i = (int)(ch - ' ')) >= 95 || i < 0)
	return (KeySym)0;
    else
	return fullchar[i].xkey;
}
