/*
    Copyright (C) 2006 by OXIM TEAM

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/xpm.h>
#include <string.h>
#include "oximtool.h"
#include "gui.h"
#include "oxim.h"

#include "images/halfchar.xpm"
#include "images/fullchar.xpm"
#include "images/out_cn.xpm"
#include "images/out_tw.xpm"
#include "images/sideback_l.xpm"
#include "images/sideback.xpm"
#include "images/sideback_r.xpm"

#define XIMPreeditMask	0x00ffL
#define XIMStatusMask	0xff00L
#define X_LEADING       2
#define Y_LEADING       2

static char **last_inpb = NULL; /* 全半形狀態 */
static char **last_inp_conv = NULL; /* 繁簡狀態 */
static char last_inpn[100] = {'\0'};

static void 
win_draw(gui_t *gui, winlist_t *win, int isForce)
{
    xccore_t *xccore = (xccore_t *)win->data;
    IC *ic = xccore->ic;

    char *inpn=NULL;
    char **inpb=NULL, **inp_conv=NULL;
    char *s, buf[100];

    IM_Context_t *imc = ic->imc;
    int x, y, len, img_y;
    img_y = (win->height - 16) / 2;
    inpb = ((imc->inp_state & IM_2BYTES)) ? icon_full : icon_half;

    if (imc->inp_state & IM_OUTSIMP)
	inp_conv = icon_cn;
    else if (imc->inp_state & IM_OUTTRAD)
	inp_conv = icon_tw;

    if ((imc->inp_state & IM_CINPUT))
    {
	im_t *im = oxim_get_IMByIndex(imc->inp_num);
        if (im)
	{
            strcpy(buf, (im->aliasname ? im->aliasname : im->inpname));
            inpn = buf;
        }
	else
	{
	    inpn = "No Name";
	}
    }
    else
        inpn = "En";

    /* 狀態沒有異動, 就不重繪 */
    if (!isForce && inpb == last_inpb && inp_conv == last_inp_conv &&
	strcmp(inpn, last_inpn) == 0)
	return;

    len = strlen((char *)inpn);

    gui_winmap_change(win, 1);
    XClearWindow(gui->display, win->window);
    
    gui_Draw_Image(win, 0, 0, sideback_l);
    x = 15 ;
    y = win->height - ((win->height - win->font_size) / 2) - 1;

    int inpn_w = gui_TextEscapement(win, inpn, len);

    gui_Draw_String(win, FONT_COLOR, NO_COLOR, x+1, y+1, inpn, len);
    x += gui_Draw_String(win, LIGHT_COLOR, NO_COLOR, x, y, inpn, len);

    x += (X_LEADING * 2) + 1;
    gui_Draw_Image(win, x, img_y, inpb);
    x += 16 + (X_LEADING * 2);

    if (inp_conv)
    {
	gui_Draw_Image(win, x, img_y, inp_conv);
	x += 20;
    }

    win->width = x + 5;
    gui_Draw_Image(win, win->width - 5, 0, sideback_r);

    if (win->width + win->pos_x > gui->display_width)
    {
	win->pos_x = gui->display_width - win->width;
	XMoveWindow(gui->display, win->window, win->pos_x, win->pos_y);
    }
    XResizeWindow(gui->display, win->window, win->width, win->height);

    /* 紀錄最後狀態 */
    last_inpb = inpb;
    last_inp_conv = inp_conv;
    strcpy(last_inpn, inpn);
}

static void
gui_status_draw(winlist_t *win)
{
    xccore_t *xccore = (xccore_t *)win->data;
    IC *ic = xccore->ic;
    inp_state_t inp_state = (ic) ? ic->imc->inp_state : 0;
    if (((win->winmode & WMODE_EXIT)) || !ic || (!(inp_state & IM_CINPUT) && !(inp_state & IM_2BYTES)) || gui->xcin_style || (ic->ic_rec.input_style & XIMPreeditMask) == XIMPreeditNothing || (ic->ic_rec.input_style & XIMPreeditMask) == XIMPreeditArea)
    {
	gui_winmap_change(win, 0);
	last_inpb = last_inp_conv = NULL;
	last_inpn[0] = '\0';
	return;
    }

    win_draw(gui, win, (!(win->winmode & WMODE_MAP) ? True : False));
}

static void ClientMessageProcress(XEvent *event)
{
    int cmd = event->xclient.data.l[0];

    switch (cmd)
    {
    case OXIM_CMD_RELOAD: /* 重新載入 config */
	{
	    oxim_reload();
	}
	break;

    case OXIM_CMD_SET_LOCATION: /* 設定游標位置 */
	{
	    Window w = event->xclient.data.l[1]; /* focus window */
	    int x = event->xclient.data.l[2]; /* X */
	    int y = event->xclient.data.l[3]; /* Y */
	    ic_set_location(w, x, y);
	}
	break;
    }
}

static void StatusEventProcess(winlist_t *win, XEvent *event)
{
    switch (event->type)
    {
	case Expose:
	    if (event->xexpose.count == 0)
	    {
		win_draw(gui, win, True); /* 強迫重繪 */
	    }
	    break;
	case MotionNotify: /* Mouse Move */
	    if (event->xbutton.x < 12)
	    {
		XDefineCursor(gui->display, win->window, gui_cursor_move);
	    }
	    else
	    {
		XDefineCursor(gui->display, win->window, gui_cursor_arrow);
	    }
	    break;
	case ButtonPress:
	    if (event->xbutton.button == Button1 && event->xbutton.x < 12)
	    {
		gui_move_window(win);
	    }
	    break;
	case ClientMessage:
	    ClientMessageProcress(event);
	    break;
    }
}

void gui_status_init(void)
{
    winlist_t *win = NULL;

    if (gui->status_win)
    {
	win = gui->status_win;
    }
    else
    {
	win = gui_new_win();
    }

    unsigned int font_size = atoi(oxim_get_config(StatusFontSize));

    if (font_size < 12 || font_size > 18)
        font_size = 16;

    win->width    = 1;
    win->height   = 24;
    win->font_size = font_size;
    win->pos_x = gui->display_width - 200;
    win->pos_y = gui->display_height - 64;

    win->win_draw_func = gui_status_draw;
    win->win_event_func = StatusEventProcess;

    Pixmap pix;
    XpmCreatePixmapFromData(gui->display, win->window, sideback, &pix, NULL, NULL);
    XSetWindowAttributes win_attr;
    win_attr.background_pixmap = pix;
    win_attr.override_redirect = True;
    XChangeWindowAttributes(gui->display, win->window,
			CWBackPixmap|CWOverrideRedirect, &win_attr);
    XFreePixmap(gui->display, pix);

    Atom atom = XInternAtom(gui->display, OXIM_ATOM, False);
    XSetSelectionOwner(gui->display, atom, win->window, CurrentTime);

    gui->status_win = win;
}
